import typing
from abc import ABC, abstractmethod
from functools import cached_property

from fntypes.option import Nothing, Some

from telegrinder.api.api import API
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.bot.dispatch.handler.func import Func, FuncHandler
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.process import process_inner
from telegrinder.bot.dispatch.return_manager.abc import ABCReturnManager
from telegrinder.bot.dispatch.view.abc import ABCStateView, ABCView
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.model import Model
from telegrinder.msgspec_utils import Option
from telegrinder.tools.error_handler.error_handler import ABCErrorHandler, ErrorHandler
from telegrinder.types.objects import Update


def get_event_model_class[Event: BaseCute](
    view: "BaseView[Event] | type[BaseView[Event]]",
) -> Option[type[Event]]:
    view_class = view if isinstance(view, typing.Type) else view.__class__
    for base in view.__class__.__bases__ + (view_class,):
        if "__orig_bases__" not in base.__dict__:
            continue

        for orig_base in base.__dict__["__orig_bases__"]:
            origin_base = typing.get_origin(orig_base) or orig_base
            if not isinstance(origin_base, type) and not issubclass(origin_base, object):
                continue

            for generic_type in typing.get_args(orig_base):
                orig_generic_type = typing.get_origin(generic_type) or generic_type
                if isinstance(orig_generic_type, type) and issubclass(orig_generic_type, BaseCute):
                    return Some(generic_type)

    return Nothing()


class BaseView[Event: BaseCute](ABCView):
    def __init__(self) -> None:
        self.handlers: list[ABCHandler[Event]] = []
        self.middlewares: list[ABCMiddleware[Event]] = []
        self.return_manager: ABCReturnManager[Event] | None = None
        self._auto_rules: ABCRule | None = None

    @property
    def auto_rules(self) -> tuple[ABCRule] | tuple[()]:
        return (self._auto_rules,) if self._auto_rules else ()

    @auto_rules.setter
    def auto_rules(self, value: ABCRule | None) -> None:
        """Example usage:

        ```python
        view.auto_rules = Rule1() & Rule2() | Rule3() & Rule4()
        view.auto_rules  # (<OrRule>,)

        view.auto_rules = None
        view.auto_rules # ()
        ```
        """
        self._auto_rules = value

    @staticmethod
    def get_raw_event(update: Update) -> Option[Model]:
        return getattr(update, update.update_type.value)

    @cached_property
    def event_model_class(self) -> Option[type[Event]]:
        return get_event_model_class(self)

    @typing.overload
    @classmethod
    def to_handler[**P, R](
        cls,
        *rules: ABCRule,
    ) -> typing.Callable[
        [Func[P, R]],
        FuncHandler[Event, Func[P, R], ErrorHandler[Event]],
    ]: ...

    @typing.overload
    @classmethod
    def to_handler[**P, Dataclass, R](
        cls,
        *rules: ABCRule,
        dataclass: type[Dataclass],
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]],
        FuncHandler[Dataclass, Func[P, R], ErrorHandler[Dataclass]],
    ]: ...

    @typing.overload
    @classmethod
    def to_handler[**P, ErrorHandlerT: ABCErrorHandler, R](
        cls,
        *rules: ABCRule,
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[Event, Func[P, R], ErrorHandlerT]]: ...

    @typing.overload
    @classmethod
    def to_handler[**P, Dataclass, ErrorHandlerT: ABCErrorHandler, R](
        cls,
        *rules: ABCRule,
        dataclass: type[Dataclass],
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[Dataclass, Func[P, R], ErrorHandlerT]]: ...

    @classmethod
    def to_handler(
        cls,
        *rules: ABCRule,
        dataclass: type[typing.Any] | None = None,
        error_handler: ABCErrorHandler | None = None,
        final: bool = True,
    ) -> typing.Callable[..., typing.Any]:
        def wrapper(func):
            return FuncHandler(
                func,
                list(rules),
                final=final,
                dataclass=dataclass,
                error_handler=error_handler or ErrorHandler(),
            )

        return wrapper

    @typing.overload
    def __call__[**P, R](
        self,
        *rules: ABCRule,
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]],
        FuncHandler[Event, Func[P, R], ErrorHandler[Event]],
    ]: ...

    @typing.overload
    def __call__[**P, Dataclass, R](
        self,
        *rules: ABCRule,
        dataclass: type[Dataclass],
        final: bool = True,
    ) -> typing.Callable[
        [Func[P, R]],
        FuncHandler[Dataclass, Func[P, R], ErrorHandler[Dataclass]],
    ]: ...

    @typing.overload
    def __call__[**P, ErrorHandlerT: ABCErrorHandler, R](
        self,
        *rules: ABCRule,
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[Event, Func[P, R], ErrorHandlerT]]: ...

    def __call__[**P, R](
        self,
        *rules: ABCRule,
        dataclass: type[typing.Any] | None = None,
        error_handler: ABCErrorHandler | None = None,
        final: bool = True,
    ) -> typing.Callable[..., typing.Any]:
        def wrapper(func: typing.Callable[..., typing.Any]):
            func_handler = FuncHandler(
                func,
                [*self.auto_rules, *rules],
                final=final,
                dataclass=dataclass,
                error_handler=error_handler or ErrorHandler(),
            )
            self.handlers.append(func_handler)
            return func_handler

        return wrapper

    def register_middleware[Middleware: ABCMiddleware](self, *args: typing.Any, **kwargs: typing.Any):
        def wrapper(cls: type[Middleware]) -> type[Middleware]:
            self.middlewares.append(cls(*args, **kwargs))
            return cls

        return wrapper

    async def check(self, event: Update) -> bool:
        match self.get_raw_event(event):
            case Some(e) if issubclass(
                self.event_model_class.expect(
                    "{!r} has no event model class in generic.".format(self.__class__.__qualname__),
                ),
                e.__class__,
            ) and (self.handlers or self.middlewares):
                return True
            case _:
                return False

    async def process(self, event: Update, api: API, context: Context) -> bool:
        return await process_inner(
            api,
            self.event_model_class.unwrap().from_update(
                update=self.get_raw_event(event).unwrap(),
                bound_api=api,
            ),
            event,
            context,
            self.middlewares,
            self.handlers,
            self.return_manager,
        )

    def load(self, external: typing.Self, /) -> None:
        self.handlers.extend(external.handlers)
        self.middlewares.extend(external.middlewares)


class BaseStateView[Event: BaseCute](ABCStateView[Event], BaseView[Event], ABC):
    @classmethod
    @abstractmethod
    def get_state_key(cls, event: Event) -> int | None:
        pass


__all__ = ("ABCStateView", "ABCView", "BaseStateView", "BaseView")



from fntypes import Nothing, Option, Some


def from_optional[Value](value: Value | None, /) -> Option[Value]:
    return Some(value) if value is not None else Nothing()


__all__ = ("from_optional",)



from __future__ import annotations

import dataclasses

import typing_extensions as typing
from fntypes import Nothing, Option, Some
from vbml.patcher import Patcher

from telegrinder.api.api import API, HTTPClient
from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import ErrorHandlerT, Func, FuncHandler
from telegrinder.bot.dispatch.middleware.abc import run_middleware
from telegrinder.bot.dispatch.middleware.global_middleware import GlobalMiddleware
from telegrinder.bot.dispatch.view.abc import ABCView
from telegrinder.bot.dispatch.view.box import (
    CallbackQueryView,
    ChatJoinRequestView,
    ChatMemberView,
    InlineQueryView,
    MessageView,
    PreCheckoutQueryView,
    RawEventView,
    ViewBox,
)
from telegrinder.modules import logger
from telegrinder.tools.error_handler.error_handler import ErrorHandler
from telegrinder.tools.global_context import TelegrinderContext
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.cute_types.base import BaseCute
    from telegrinder.bot.cute_types.update import UpdateCute
    from telegrinder.bot.rules.abc import ABCRule

T = typing.TypeVar("T", default=typing.Any)
R = typing.TypeVar("R", covariant=True, default=typing.Any)
Event = typing.TypeVar("Event", bound="BaseCute")
P = typing.ParamSpec("P", default=...)

DEFAULT_DATACLASS: typing.Final[type[Update]] = Update


@dataclasses.dataclass(repr=False, kw_only=True)
class Dispatch(
    ABCDispatch,
    ViewBox[
        CallbackQueryView,
        ChatJoinRequestView,
        ChatMemberView,
        InlineQueryView,
        MessageView,
        PreCheckoutQueryView,
        RawEventView,
    ],
    typing.Generic[
        HTTPClient,
        CallbackQueryView,
        ChatJoinRequestView,
        ChatMemberView,
        InlineQueryView,
        MessageView,
        PreCheckoutQueryView,
        RawEventView,
    ],
):
    _global_context: TelegrinderContext = dataclasses.field(
        init=False,
        default_factory=TelegrinderContext,
    )
    global_middleware: "GlobalMiddleware" = dataclasses.field(
        default_factory=lambda: GlobalMiddleware(),
    )

    def __repr__(self) -> str:
        return "Dispatch(%s)" % ", ".join(f"{k}={v!r}" for k, v in self.get_views().items())

    @property
    def global_context(self) -> TelegrinderContext:
        return self._global_context

    @property
    def patcher(self) -> Patcher:
        """Alias `patcher` to get `vbml.Patcher` from the global context."""
        return self.global_context.vbml_patcher

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandler[UpdateCute]]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        dataclass: type[T],
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandler[T]]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        update_type: UpdateType,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandler[UpdateCute]]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        dataclass: type[T],
        update_type: UpdateType,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandler[T]]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandlerT]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        update_type: UpdateType,
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler["UpdateCute", Func[P, R], ErrorHandlerT]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        dataclass: type[T],
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[T, Func[P, R], ErrorHandlerT]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        dataclass: type[T],
        update_type: UpdateType,
        error_handler: ErrorHandlerT,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[T, Func[P, R], ErrorHandlerT]]: ...

    @typing.overload
    def handle(
        self,
        *rules: "ABCRule",
        update_type: UpdateType | None = None,
        dataclass: type[T] = DEFAULT_DATACLASS,
        error_handler: typing.Literal[None] = None,
        final: bool = True,
    ) -> typing.Callable[[Func[P, R]], FuncHandler[T, Func[P, R], ErrorHandler[T]]]: ...

    def handle(
        self,
        *rules: "ABCRule",
        update_type: UpdateType | None = None,
        dataclass: type[typing.Any] = DEFAULT_DATACLASS,
        error_handler: ErrorHandlerT | None = None,
        final: bool = True,
    ) -> typing.Callable[..., typing.Any]:
        def wrapper(func):
            handler = FuncHandler(
                func,
                list(rules),
                final=final,
                dataclass=dataclass,
                error_handler=error_handler or ErrorHandler(),
                update_type=update_type,
            )
            self.raw_event.handlers.append(handler)
            return handler

        return wrapper

    async def feed(self, event: Update, api: API[HTTPClient]) -> bool:
        logger.debug(
            "Processing update (update_id={}, update_type={!r})",
            event.update_id,
            event.update_type.name,
        )
        context = Context(raw_update=event)

        if (
            await run_middleware(
                self.global_middleware.pre,
                api,
                event,  # type: ignore
                raw_event=event,
                ctx=context,
                adapter=self.global_middleware.adapter,
            )
            is False
        ):
            return False

        for view in self.get_views().values():
            if await view.check(event):
                logger.debug(
                    "Update (update_id={}, update_type={!r}) matched view {!r}.",
                    event.update_id,
                    event.update_type.name,
                    view,
                )
                if await view.process(event, api, context):
                    return True

        await run_middleware(
            self.global_middleware.post,
            api,
            event,
            raw_event=event,
            ctx=context,
            adapter=self.global_middleware.adapter,
        )

        return False

    def load(self, external: typing.Self) -> None:
        views_external = external.get_views()

        for name, view in self.get_views().items():
            assert name in views_external, f"View {name!r} is undefined in external dispatch."
            view.load(views_external[name])
            setattr(external, name, view)

        self.global_middleware.filters.difference_update(external.global_middleware.filters)

    def get_view(self, of_type: type[T]) -> Option[T]:
        for view in self.get_views().values():
            if isinstance(view, of_type):
                return Some(view)
        return Nothing()

    def get_views(self) -> dict[str, ABCView]:
        """Get all views."""
        return {name: view for name, view in self.__dict__.items() if isinstance(view, ABCView)}

    __call__ = handle


__all__ = ("Dispatch",)



import typing

from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.handler.abc import ABCHandler

from .short_state import ShortState


class WaiterActions[Event: BaseCute](typing.TypedDict):
    on_miss: typing.NotRequired[ABCHandler[Event]]
    on_drop: typing.NotRequired[typing.Callable[[ShortState[Event]], None]]


__all__ = ("WaiterActions",)



import dataclasses
import inspect
import typing

from fntypes.error import UnwrapError
from fntypes.result import Error, Ok, Result

from telegrinder.bot.dispatch.context import Context
from telegrinder.modules import logger
from telegrinder.node.base import (
    ComposeError,
    IsNode,
    Name,
    NodeImpersonation,
    NodeScope,
    NodeType,
    unwrap_node,
)
from telegrinder.tools.magic import join_dicts, magic_bundle

type AsyncGenerator = typing.AsyncGenerator[typing.Any, None]

CONTEXT_STORE_NODES_KEY = "_node_ctx"
GLOBAL_VALUE_KEY = "_value"


def get_scope(node: type[NodeType], /) -> NodeScope | None:
    return getattr(node, "scope", None)


async def compose_node(
    node: type[NodeType],
    linked: dict[type[typing.Any], typing.Any],
    data: dict[type[typing.Any], typing.Any] | None = None,
) -> "NodeSession":
    subnodes = node.get_subnodes()
    kwargs = magic_bundle(node.compose, join_dicts(subnodes, linked))

    # Linking data via typebundle
    if data:
        kwargs.update(magic_bundle(node.compose, data, typebundle=True))

    if node.is_generator():
        generator = typing.cast(AsyncGenerator, node.compose(**kwargs))
        value = await generator.asend(None)
    else:
        generator = None
        value = node.compose(**kwargs)
        if inspect.isawaitable(value):
            value = await value

    return NodeSession(node, value, subnodes={}, generator=generator)


async def compose_nodes(
    nodes: typing.Mapping[str, IsNode | NodeImpersonation],
    ctx: Context,
    data: dict[type[typing.Any], typing.Any] | None = None,
) -> Result["NodeCollection", ComposeError]:
    logger.debug("Composing nodes: ({})...", " ".join(f"{k}={v!r}" for k, v in nodes.items()))

    data = {Context: ctx} | (data or {})
    parent_nodes = dict[IsNode, NodeSession]()
    event_nodes: dict[IsNode, NodeSession] = ctx.get_or_set(CONTEXT_STORE_NODES_KEY, {})
    unwrapped_nodes = {(key, n := node.as_node()): unwrap_node(n) for key, node in nodes.items()}

    for (parent_node_name, parent_node_t), linked_nodes in unwrapped_nodes.items():
        local_nodes = dict[type[NodeType], NodeSession]()
        subnodes = {}
        data[Name] = parent_node_name

        for node_t in linked_nodes:
            scope = get_scope(node_t)

            if scope is NodeScope.PER_EVENT and node_t in event_nodes:
                local_nodes[node_t] = event_nodes[node_t]
                continue
            elif scope is NodeScope.GLOBAL and hasattr(node_t, GLOBAL_VALUE_KEY):
                local_nodes[node_t] = NodeSession(node_t, getattr(node_t, GLOBAL_VALUE_KEY), {})
                continue

            subnodes |= {
                k: session.value for k, session in (local_nodes | event_nodes).items() if k not in subnodes
            }
            try:
                local_nodes[node_t] = await compose_node(node_t, linked=subnodes, data=data)
            except (ComposeError, UnwrapError) as exc:
                for t, local_node in local_nodes.items():
                    if get_scope(t) is NodeScope.PER_CALL:
                        await local_node.close()
                return Error(ComposeError(f"Cannot compose {node_t!r}, error: {str(exc)}"))

            if scope is NodeScope.PER_EVENT:
                event_nodes[node_t] = local_nodes[node_t]
            elif scope is NodeScope.GLOBAL:
                setattr(node_t, GLOBAL_VALUE_KEY, local_nodes[node_t].value)

        parent_nodes[parent_node_t] = local_nodes[parent_node_t]

    return Ok(NodeCollection({k: parent_nodes[t] for k, t in unwrapped_nodes}))


@dataclasses.dataclass(slots=True, repr=False)
class NodeSession:
    node_type: type[NodeType] | None
    value: typing.Any
    subnodes: dict[str, typing.Self]
    generator: typing.AsyncGenerator[typing.Any, typing.Any | None] | None = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.value!r}" + (" (ACTIVE)>" if self.generator else ">")

    async def close(
        self,
        with_value: typing.Any | None = None,
        scopes: tuple[NodeScope, ...] = (NodeScope.PER_CALL,),
    ) -> None:
        if self.node_type and getattr(self.node_type, "scope", None) not in scopes:
            return

        for subnode in self.subnodes.values():
            await subnode.close(scopes=scopes)

        if self.generator is None:
            return
        try:
            logger.debug("Closing session for node {!r}...", self.node_type)
            await self.generator.asend(with_value)
        except StopAsyncIteration:
            self.generator = None


class NodeCollection:
    __slots__ = ("sessions", "_values")

    def __init__(self, sessions: dict[str, NodeSession]) -> None:
        self.sessions = sessions
        self._values: dict[str, typing.Any] = {}

    def __repr__(self) -> str:
        return "<{}: sessions={!r}>".format(self.__class__.__name__, self.sessions)

    @property
    def values(self) -> dict[str, typing.Any]:
        if self._values.keys() == self.sessions.keys():
            return self._values

        for name, session in self.sessions.items():
            if name not in self._values:
                self._values[name] = session.value

        return self._values

    async def close_all(
        self,
        with_value: typing.Any | None = None,
        scopes: tuple[NodeScope, ...] = (NodeScope.PER_CALL,),
    ) -> None:
        for session in self.sessions.values():
            await session.close(with_value, scopes=scopes)


__all__ = ("NodeCollection", "NodeSession", "compose_node", "compose_nodes")



import re

import vbml

from telegrinder.tools.global_context import GlobalContext, ctx_var


class TelegrinderContext(GlobalContext):
    """Basic type-hinted telegrinder context with context name `"telegrinder"`.

    You can use this class or GlobalContext:
    ```
    from telegrinder.tools.global_context import GlobalContext, TelegrinderContext

    ctx1 = TelegrinderContext()
    ctx2 = GlobalContext("telegrinder")  # same, but without the type-hints
    assert ctx1 == ctx2  # ok
    ```
    """

    __ctx_name__ = "telegrinder"

    vbml_pattern_flags: re.RegexFlag | None = None
    vbml_patcher: vbml.Patcher = ctx_var(default=vbml.Patcher(), frozen=True)


__all__ = ("TelegrinderContext",)



import typing

from fntypes.result import Ok

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.node.base import ComposeError, FactoryNode, Node
from telegrinder.node.composer import CONTEXT_STORE_NODES_KEY, GLOBAL_VALUE_KEY, compose_node, compose_nodes
from telegrinder.node.scope import NodeScope, per_call
from telegrinder.types.objects import Update


@per_call
class _Either(FactoryNode):
    """Represents a node that either to compose `left` or `right` nodes.

    For example:
    ```python
    # ScalarNode `Integer` -> int
    # ScalarNode `Float` -> float

    Number = Either[Integer, Float]  # using a type alias just as an example

    def number_to_int(number: Number) -> int:
        return int(number)
    ```
    """

    nodes: tuple[type[Node], type[Node] | None]

    def __class_getitem__(cls, node: type[Node] | tuple[type[Node], type[Node]], /):
        nodes = (node, None) if not isinstance(node, tuple) else node
        assert len(nodes) == 2, "Node `Either` must have at least two nodes."
        return cls(nodes=nodes)

    @classmethod
    async def compose(cls, api: API, update: Update, context: Context) -> typing.Any | None:
        data = {API: api, Update: update, Context: context}
        node_ctx = context.get_or_set(CONTEXT_STORE_NODES_KEY, {})

        for node in cls.nodes:
            if node is None:
                return None

            if node.scope is NodeScope.PER_EVENT and node in node_ctx:
                return node_ctx[node].value
            elif node.scope is NodeScope.GLOBAL and hasattr(node, GLOBAL_VALUE_KEY):
                return getattr(node, GLOBAL_VALUE_KEY)

            subnodes = node.as_node().get_subnodes()
            match await compose_nodes(subnodes, context, data):
                case Ok(col):
                    try:
                        session = await compose_node(
                            node=node,
                            linked={
                                typing.cast(type, n): col.sessions[name].value for name, n in subnodes.items()
                            },
                            data=data,
                        )
                    except ComposeError:
                        continue

                    if node.scope is NodeScope.PER_EVENT:
                        node_ctx[node] = session
                    elif node.scope is NodeScope.GLOBAL:
                        setattr(node, GLOBAL_VALUE_KEY, session.value)

                    return session.value

        raise ComposeError("Cannot compose either nodes: {}.".format(", ".join(repr(n) for n in cls.nodes)))


if typing.TYPE_CHECKING:
    type Either[Left, Right: typing.Any | None] = Left | Right
    type Optional[Left] = Either[Left, None]
else:
    Either = _Either
    Optional = type("Optional", (Either,), {})


__all__ = ("Either", "Optional")



import typing

from .abc import ABCRule, Context, UpdateCute, check_rule


class If(ABCRule):
    def __init__(self, condition: ABCRule) -> None:
        self.conditions = [condition]

    async def check(self, update: UpdateCute, ctx: Context) -> bool:
        for condition in self.conditions[:-1]:
            if not await check_rule(update.api, condition, update, ctx):
                return True
        return await check_rule(update.api, self.conditions[-1], update, ctx)

    def then(self, condition: ABCRule) -> typing.Self:
        self.conditions.append(condition)
        return self



import asyncio
import dataclasses
import datetime
import typing

type CoroutineTask[T] = typing.Coroutine[typing.Any, typing.Any, T]
type CoroutineFunc[**P, T] = typing.Callable[P, CoroutineTask[T]]
type Task[**P, T] = CoroutineFunc[P, T] | CoroutineTask[T] | DelayedTask[typing.Callable[P, CoroutineTask[T]]]


def run_tasks(
    tasks: list[CoroutineTask[typing.Any]],
    /,
) -> None:
    loop = asyncio.get_event_loop()
    while tasks:
        loop.run_until_complete(tasks.pop(0))


def to_coroutine_task[**P, T](task: Task[P, T], /) -> CoroutineTask[T]:
    if asyncio.iscoroutinefunction(task) or isinstance(task, DelayedTask):
        task = task()
    elif not asyncio.iscoroutine(task):
        raise TypeError("Task should be coroutine or coroutine function.")
    return task


@dataclasses.dataclass(slots=True)
class DelayedTask[Handler: CoroutineFunc[..., typing.Any]]:
    handler: Handler
    seconds: float | datetime.timedelta
    repeat: bool = dataclasses.field(default=False, kw_only=True)
    _cancelled: bool = dataclasses.field(default=False, init=False, repr=False)

    @property
    def is_cancelled(self) -> bool:
        return self._cancelled

    @property
    def delay(self) -> float:
        return self.seconds if isinstance(self.seconds, int | float) else self.seconds.total_seconds()

    async def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        while not self.is_cancelled:
            await asyncio.sleep(self.delay)
            if self.is_cancelled:
                break
            try:
                await self.handler(*args, **kwargs)
            finally:
                if not self.repeat:
                    break

    def cancel(self) -> None:
        if not self._cancelled:
            self._cancelled = True


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class Lifespan:
    startup_tasks: list[CoroutineTask[typing.Any]] = dataclasses.field(default_factory=lambda: [])
    shutdown_tasks: list[CoroutineTask[typing.Any]] = dataclasses.field(default_factory=lambda: [])

    def on_startup[**P, T](self, task: Task[P, T], /) -> Task[P, T]:
        self.startup_tasks.append(to_coroutine_task(task))
        return task

    def on_shutdown[**P, T](self, task: Task[P, T], /) -> Task[P, T]:
        self.shutdown_tasks.append(to_coroutine_task(task))
        return task

    def start(self) -> None:
        run_tasks(self.startup_tasks)

    def shutdown(self) -> None:
        run_tasks(self.shutdown_tasks)

    def __enter__(self) -> None:
        self.start()

    def __exit__(self) -> None:
        self.shutdown()

    async def __aenter__(self) -> None:
        for task in self.startup_tasks:
            await task

    async def __aexit__(self, *args) -> None:
        for task in self.shutdown_tasks:
            await task

    def __add__(self, other: typing.Self, /) -> typing.Self:
        return self.__class__(
            startup_tasks=self.startup_tasks + other.startup_tasks,
            shutdown_tasks=self.shutdown_tasks + other.shutdown_tasks,
        )


__all__ = (
    "CoroutineTask",
    "DelayedTask",
    "Lifespan",
    "run_tasks",
    "to_coroutine_task",
)



from .api import API
from .error import APIError, APIServerError, InvalidTokenError
from .response import APIResponse
from .token import Token

__all__ = (
    "API",
    "APIError",
    "APIResponse",
    "APIServerError",
    "InvalidTokenError",
    "Token",
)



import typing

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.dispatch.return_manager.message import MessageReturnManager
from telegrinder.bot.dispatch.view.base import BaseStateView
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Update

MessageUpdateType: typing.TypeAlias = typing.Literal[
    UpdateType.MESSAGE,
    UpdateType.BUSINESS_MESSAGE,
    UpdateType.CHANNEL_POST,
    UpdateType.EDITED_BUSINESS_MESSAGE,
    UpdateType.EDITED_CHANNEL_POST,
    UpdateType.EDITED_MESSAGE,
]


class MessageView(BaseStateView[MessageCute]):
    def __init__(self, *, update_type: MessageUpdateType | None = None) -> None:
        super().__init__()
        self.update_type = update_type
        self.return_manager = MessageReturnManager()

    def __repr__(self) -> str:
        return "<{}: {!r}>".format(
            self.__class__.__name__,
            "any message update" if self.update_type is None else self.update_type.value,
        )

    @classmethod
    def get_state_key(cls, event: MessageCute) -> int | None:
        return event.chat_id

    async def check(self, event: Update) -> bool:
        return not (
            self.update_type is not None
            and self.update_type != event.update_type
            or not await super().check(event)
        )


__all__ = ("MessageView",)



import dataclasses
import typing

from fntypes.option import Nothing, Option, Some

from telegrinder.api.api import API
from telegrinder.bot.cute_types import CallbackQueryCute, ChatJoinRequestCute, MessageCute, PreCheckoutQueryCute
from telegrinder.node.base import ComposeError, DataNode, scalar_node
from telegrinder.node.polymorphic import Polymorphic, impl
from telegrinder.types.objects import Chat, Message, User


@dataclasses.dataclass(kw_only=True, slots=True)
class Source(Polymorphic, DataNode):
    api: API
    from_user: User
    chat: Option[Chat] = dataclasses.field(default_factory=Nothing)
    thread_id: Option[int] = dataclasses.field(default_factory=Nothing)

    @impl
    def compose_message(cls, message: MessageCute) -> typing.Self:
        return cls(
            api=message.ctx_api,
            from_user=message.from_user,
            chat=Some(message.chat),
            thread_id=message.message_thread_id,
        )

    @impl
    def compose_callback_query(cls, callback_query: CallbackQueryCute) -> typing.Self:
        return cls(
            api=callback_query.ctx_api,
            from_user=callback_query.from_user,
            chat=callback_query.chat,
            thread_id=callback_query.message_thread_id,
        )

    @impl
    def compose_chat_join_request(cls, chat_join_request: ChatJoinRequestCute) -> typing.Self:
        return cls(
            api=chat_join_request.ctx_api,
            from_user=chat_join_request.from_user,
            chat=Some(chat_join_request.chat),
            thread_id=Nothing(),
        )

    @impl
    def compose_pre_checkout_query(cls, pre_checkout_query: PreCheckoutQueryCute) -> typing.Self:
        return cls(
            api=pre_checkout_query.ctx_api,
            from_user=pre_checkout_query.from_user,
            chat=Nothing(),
            thread_id=Nothing(),
        )

    async def send(self, text: str, **kwargs: typing.Any) -> Message:
        result = await self.api.send_message(
            chat_id=self.chat.map_or(self.from_user.id, lambda chat: chat.id).unwrap(),
            message_thread_id=self.thread_id.unwrap_or_none(),
            text=text,
            **kwargs,
        )
        return result.unwrap()


@scalar_node
class ChatSource:
    @classmethod
    def compose(cls, source: Source) -> Chat:
        return source.chat.expect(ComposeError("Source has no chat."))


@scalar_node
class UserSource:
    @classmethod
    def compose(cls, source: Source) -> User:
        return source.from_user


@scalar_node
class UserId:
    @classmethod
    def compose(cls, user: UserSource) -> int:
        return user.id


__all__ = ("ChatSource", "Source", "UserId", "UserSource")



import abc
import typing

from telegrinder.tools.adapter.event import EventAdapter
from telegrinder.types.objects import Message as MessageEvent

from .abc import ABCRule, CheckResult, Message


class MessageRule(ABCRule[Message], abc.ABC, adapter=EventAdapter(MessageEvent, Message)):
    @abc.abstractmethod
    def check(self, *args: typing.Any, **kwargs: typing.Any) -> CheckResult: ...


__all__ = ("MessageRule",)



import keyword
import typing

import msgspec
from fntypes.co import Nothing, Result, Some

from telegrinder.msgspec_utils import decoder, encoder, struct_as_dict

if typing.TYPE_CHECKING:
    from telegrinder.api.error import APIError

MODEL_CONFIG: typing.Final[dict[str, typing.Any]] = {
    "dict": True,
    "rename": {kw + "_": kw for kw in keyword.kwlist},
}
UNSET = typing.cast(typing.Any, msgspec.UNSET)
"""Docs: https://jcristharif.com/msgspec/api.html#unset

During decoding, if a field isn't explicitly set in the model,
the default value of `UNSET` will be set instead. This lets downstream
consumers determine whether a field was left unset, or explicitly set a value."""


def full_result[T](
    result: Result[msgspec.Raw, "APIError"],
    full_t: type[T],
) -> Result[T, "APIError"]:
    return result.map(lambda v: decoder.decode(v, type=full_t))


def is_none(value: typing.Any, /) -> typing.TypeGuard[None | Nothing]:
    return value is None or isinstance(value, Nothing)


def get_params(params: dict[str, typing.Any]) -> dict[str, typing.Any]:
    validated_params = {}
    for k, v in (
        *params.pop("other", {}).items(),
        *params.items(),
    ):
        if isinstance(v, Proxy):
            v = v.get()
        if k == "self" or is_none(v):
            continue
        validated_params[k] = v.unwrap() if isinstance(v, Some) else v
    return validated_params


if typing.TYPE_CHECKING:

    @typing.overload
    def field(*, name: str | None = ...) -> typing.Any: ...

    @typing.overload
    def field(*, default: typing.Any, name: str | None = ...) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default_factory: typing.Callable[[], typing.Any],
        name: str | None = None,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = ...,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default: typing.Any,
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = ...,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default_factory: typing.Callable[[], typing.Any],
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = None,
    ) -> typing.Any: ...

    def field(
        *,
        default=...,
        default_factory=...,
        name=...,
        converter=...,
    ) -> typing.Any: ...

    class From[T]:
        def __new__(cls, _: T, /) -> typing.Any: ...
else:
    from msgspec import field as _field

    type From[T] = T

    def field(**kwargs):
        kwargs.pop("converter", None)
        return _field(**kwargs)


@typing.dataclass_transform(field_specifiers=(field,))
class Model(msgspec.Struct, **MODEL_CONFIG):
    if not typing.TYPE_CHECKING:

        def __post_init__(self):
            for field in self.__struct_fields__:
                if is_none(getattr(self, field)):
                    setattr(self, field, UNSET)

        def __getattribute__(self, name, /):
            val = super().__getattribute__(name)
            return Nothing() if val is UNSET else val

    @classmethod
    def from_data[**P, T](cls: typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        return decoder.convert(msgspec.structs.asdict(cls(*args, **kwargs)), type=cls)  # type: ignore

    @classmethod
    def from_dict(cls, obj: dict[str, typing.Any], /) -> typing.Self:
        return decoder.convert(obj, type=cls)

    @classmethod
    def from_raw(cls, raw: str | bytes, /) -> typing.Self:
        return decoder.decode(raw, type=cls)

    def _to_dict(
        self,
        dct_name: str,
        exclude_fields: set[str],
        full: bool,
    ) -> dict[str, typing.Any]:
        if dct_name not in self.__dict__:
            self.__dict__[dct_name] = (
                struct_as_dict(self)
                if not full
                else encoder.to_builtins(self.to_dict(exclude_fields=exclude_fields), order="deterministic")
            )

        if not exclude_fields:
            return self.__dict__[dct_name]

        return {key: value for key, value in self.__dict__[dct_name].items() if key not in exclude_fields}

    def to_raw(self) -> str:
        return encoder.encode(self)

    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        """:param exclude_fields: Model field names to exclude from the dictionary representation of this model.
        :return: A dictionary representation of this model.
        """
        return self._to_dict("model_as_dict", exclude_fields or set(), full=False)

    def to_full_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        """:param exclude_fields: Model field names to exclude from the dictionary representation of this model.
        :return: A dictionary representation of this model including all models, structs, custom types.
        """
        return self._to_dict("model_as_full_dict", exclude_fields or set(), full=True)


class Proxy[T]:
    def __init__(self, cfg: "_ProxiedDict[T]", key: str) -> None:
        self.key = key
        self.cfg = cfg

    def get(self) -> typing.Any | None:
        return self.cfg._defaults.get(self.key)


class _ProxiedDict[T]:
    def __init__(self, tp: type[T]) -> None:
        self.type = tp
        self._defaults = {}

    def __setattribute__(self, name: str, value: typing.Any, /) -> None:
        self._defaults[name] = value

    def __getitem__(self, key: str, /) -> None:
        return Proxy(self, key)  # type: ignore

    def __setitem__(self, key: str, value: typing.Any, /) -> None:
        self._defaults[key] = value


if typing.TYPE_CHECKING:

    def ProxiedDict[T](typed_dct: type[T]) -> T | _ProxiedDict[T]:  # noqa: N802
        ...
else:
    ProxiedDict = _ProxiedDict


__all__ = (
    "MODEL_CONFIG",
    "Model",
    "ProxiedDict",
    "Proxy",
    "full_result",
    "get_params",
)



import dataclasses
from functools import cached_property

import typing_extensions as typing

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.modules import logger
from telegrinder.node.base import NodeType, get_nodes
from telegrinder.node.composer import NodeCollection, compose_nodes
from telegrinder.tools.adapter.abc import ABCAdapter
from telegrinder.tools.adapter.dataclass import DataclassAdapter
from telegrinder.tools.error_handler import ABCErrorHandler, ErrorHandler
from telegrinder.tools.magic import get_annotations, magic_bundle
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Update

from .abc import ABCHandler

if typing.TYPE_CHECKING:
    from telegrinder.bot.rules.abc import ABCRule
    from telegrinder.node.composer import NodeCollection

Function = typing.TypeVar("Function", bound="Func[..., typing.Any]")
Event = typing.TypeVar("Event")
ErrorHandlerT = typing.TypeVar("ErrorHandlerT", bound=ABCErrorHandler, default=ErrorHandler)

type Func[**Rest, Result] = typing.Callable[Rest, typing.Coroutine[typing.Any, typing.Any, Result]]


@dataclasses.dataclass(repr=False, slots=True)
class FuncHandler(ABCHandler[Event], typing.Generic[Event, Function, ErrorHandlerT]):
    function: Function
    rules: list["ABCRule"]
    adapter: ABCAdapter[Update, Event] | None = dataclasses.field(default=None, kw_only=True)
    final: bool = dataclasses.field(default=True, kw_only=True)
    dataclass: type[typing.Any] | None = dataclasses.field(default=dict[str, typing.Any], kw_only=True)
    error_handler: ErrorHandlerT = dataclasses.field(
        default_factory=lambda: typing.cast(ErrorHandlerT, ErrorHandler()),
        kw_only=True,
    )
    preset_context: Context = dataclasses.field(default_factory=lambda: Context(), kw_only=True)
    update_type: UpdateType | None = dataclasses.field(default=None, kw_only=True)

    def __post_init__(self) -> None:
        self.dataclass = typing.get_origin(self.dataclass) or self.dataclass

        if self.dataclass is not None and self.adapter is None:
            self.adapter = DataclassAdapter(self.dataclass, self.update_type)

    @property
    def __call__(self) -> Function:
        return self.function

    def __repr__(self) -> str:
        return "<{}: {}={!r} with rules={!r}, dataclass={!r}, error_handler={!r}>".format(
            self.__class__.__name__,
            "final function" if self.final else "function",
            self.function.__qualname__,
            self.rules,
            self.dataclass,
            self.error_handler,
        )

    @cached_property
    def required_nodes(self) -> dict[str, type[NodeType]]:
        return get_nodes(self.function)

    def get_name_event_param(self, event: Event) -> str | None:
        event_class = self.dataclass or event.__class__
        for k, v in get_annotations(self.function).items():
            if isinstance(v := typing.get_origin(v) or v, type) and v is event_class:
                self.func_event_param = k
                return k
        return None

    async def check(self, api: API, event: Update, ctx: Context | None = None) -> bool:
        if self.update_type is not None and self.update_type != event.update_type:
            return False

        logger.debug("Checking handler {!r}...", self)
        ctx = Context(raw_update=event) if ctx is None else ctx
        temp_ctx = ctx.copy()
        temp_ctx |= self.preset_context.copy()
        update = event

        for rule in self.rules:
            if not await check_rule(api, rule, update, temp_ctx):
                logger.debug("Rule {!r} failed!", rule)
                return False

        nodes = self.required_nodes
        node_col = None
        if nodes:
            result = await compose_nodes(nodes, ctx, data={Update: update, API: api})
            if not result:
                logger.debug(f"Cannot compose nodes for handler, error: {str(result.error)}")
                return False

            node_col = result.value
            temp_ctx |= node_col.values

        logger.debug("All checks passed for handler.")
        temp_ctx["node_col"] = node_col
        ctx |= temp_ctx
        return True

    async def run(
        self,
        api: API,
        event: Event,
        ctx: Context,
        node_col: "NodeCollection | None" = None,
    ) -> typing.Any:
        logger.debug(f"Running handler {self!r}...")

        try:
            if event_param := self.get_name_event_param(event):
                ctx = Context(**{event_param: event, **ctx})
            return await self(**magic_bundle(self.function, ctx, start_idx=0))
        except BaseException as exception:
            return await self.error_handler.run(exception, event, api, ctx)
        finally:
            if node_col := ctx.node_col:
                await node_col.close_all()


__all__ = ("FuncHandler",)



import typing
from abc import ABC, abstractmethod


class ABCLoopWrapper(ABC):
    @property
    @abstractmethod
    def is_running(self) -> bool:
        pass

    @abstractmethod
    def add_task(self, task: typing.Any, /) -> None:
        pass

    @abstractmethod
    def run_event_loop(self) -> typing.NoReturn:
        raise NotImplementedError


__all__ = ("ABCLoopWrapper",)



from __future__ import annotations

import abc
import inspect
from collections import deque
from types import AsyncGeneratorType, CodeType, resolve_bases

import typing_extensions as typing

from telegrinder.node.scope import NodeScope
from telegrinder.tools.magic import cache_magic_value, get_annotations
from telegrinder.tools.strings import to_pascal_case

if typing.TYPE_CHECKING:
    from telegrinder.node.tools.generator import generate_node
else:

    def generate_node(*args, **kwargs):
        globalns = globals()
        if "__generate_node" not in globalns:
            import telegrinder.node.tools.generator

            globals()["__generate_node"] = telegrinder.node.tools.generator.generate_node

        return globals()["__generate_node"](*args, **kwargs)


type NodeType = Node | NodeProto[typing.Any]
type IsNode = NodeType | type[NodeType]

T = typing.TypeVar("T", default=typing.Any)

ComposeResult: typing.TypeAlias = T | typing.Awaitable[T] | typing.AsyncGenerator[T, None]

UNWRAPPED_NODE_KEY = "__unwrapped_node__"


@typing.overload
def is_node(maybe_node: type[typing.Any], /) -> typing.TypeIs[type[NodeType]]: ...


@typing.overload
def is_node(maybe_node: typing.Any, /) -> typing.TypeIs[NodeType]: ...


def is_node(maybe_node: typing.Any, /) -> bool:
    if isinstance(maybe_node, typing.TypeAliasType):
        maybe_node = maybe_node.__value__
    if not isinstance(maybe_node, type):
        maybe_node = typing.get_origin(maybe_node) or maybe_node

    return (
        hasattr(maybe_node, "as_node")
        or isinstance(maybe_node, type)
        and issubclass(maybe_node, (Node, NodeProto))
        or not isinstance(maybe_node, type)
        and isinstance(maybe_node, (Node, NodeProto))
    )


@typing.overload
def as_node(maybe_node: type[typing.Any], /) -> type[NodeType]: ...


@typing.overload
def as_node(maybe_node: typing.Any, /) -> NodeType: ...


@typing.overload
def as_node(*maybe_nodes: type[typing.Any]) -> tuple[type[NodeType], ...]: ...


@typing.overload
def as_node(*maybe_nodes: typing.Any) -> tuple[NodeType, ...]: ...


@typing.overload
def as_node(*maybe_nodes: type[typing.Any] | typing.Any) -> tuple[IsNode, ...]: ...


def as_node(*maybe_nodes: typing.Any) -> typing.Any | tuple[typing.Any, ...]:
    for maybe_node in maybe_nodes:
        if not is_node(maybe_node):
            is_type = isinstance(maybe_node, type)
            raise LookupError(
                f"{'Type of' if is_type else 'Object of type'} "
                f"{maybe_node.__name__ if is_type else maybe_node.__class__.__name__!r} "
                "cannot be resolved as Node."
            )
    return maybe_nodes[0] if len(maybe_nodes) == 1 else maybe_nodes


@cache_magic_value("__nodes__")
def get_nodes(function: typing.Callable[..., typing.Any], /) -> dict[str, type[NodeType]]:
    return {k: v.as_node() for k, v in get_annotations(function).items() if is_node(v)}


@cache_magic_value("__is_generator__")
def is_generator(
    function: typing.Callable[..., typing.Any],
    /,
) -> typing.TypeGuard[AsyncGeneratorType[typing.Any, None]]:
    return inspect.isasyncgenfunction(function)


def unwrap_node(node: type[NodeType], /) -> tuple[type[NodeType], ...]:
    """Unwrap node as flattened tuple of node types in ordering required to calculate given node.

    Provides caching for passed node type.
    """
    if (unwrapped := getattr(node, UNWRAPPED_NODE_KEY, None)) is not None:
        return unwrapped

    stack = deque([(node, node.get_subnodes().values())])
    visited = list[type[NodeType]]()

    while stack:
        parent, child_nodes = stack.pop()

        if parent not in visited:
            visited.insert(0, parent)

        for child in child_nodes:
            stack.append((child, child.get_subnodes().values()))

    unwrapped = tuple(visited)
    setattr(node, UNWRAPPED_NODE_KEY, unwrapped)
    return unwrapped


class ComposeError(BaseException):
    pass


@typing.runtime_checkable
class Composable[R](typing.Protocol):
    @classmethod
    def compose(cls, *args: typing.Any, **kwargs: typing.Any) -> ComposeResult[R]: ...


class NodeImpersonation(typing.Protocol):
    @classmethod
    def as_node(cls) -> type[NodeProto[typing.Any]]: ...


class NodeComposeFunction[R](typing.Protocol):
    __name__: str
    __code__: CodeType

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> ComposeResult[R]: ...


@typing.runtime_checkable
class NodeProto[R](Composable[R], NodeImpersonation, typing.Protocol):
    @classmethod
    def get_subnodes(cls) -> dict[str, type[NodeType]]: ...

    @classmethod
    def is_generator(cls) -> bool: ...


class Node(abc.ABC):
    node: str = "node"
    scope: NodeScope = NodeScope.PER_EVENT

    @classmethod
    @abc.abstractmethod
    def compose(cls, *args, **kwargs) -> ComposeResult:
        pass

    @classmethod
    def get_subnodes(cls) -> dict[str, type[NodeType]]:
        return get_nodes(cls.compose)

    @classmethod
    def as_node(cls) -> type[typing.Self]:
        return cls

    @classmethod
    def is_generator(cls) -> bool:
        return is_generator(cls.compose)


class scalar_node[T]:  # noqa: N801
    @typing.overload
    def __new__(cls, x: NodeComposeFunction[Composable[T]], /) -> type[T]: ...

    @typing.overload
    def __new__(cls, x: NodeComposeFunction[T], /) -> type[T]: ...

    @typing.overload
    def __new__(
        cls,
        /,
        *,
        scope: NodeScope,
    ) -> typing.Callable[[NodeComposeFunction[Composable[T]] | NodeComposeFunction[T]], type[T]]: ...

    def __new__(cls, x=None, /, *, scope=NodeScope.PER_EVENT) -> typing.Any:
        def inner(node_or_func, /) -> typing.Any:
            namespace = {"node": "scalar", "scope": scope, "__module__": node_or_func.__module__}

            if isinstance(node_or_func, type):
                bases: list[type[typing.Any]] = [node_or_func]
                node_bases = resolve_bases(node_or_func.__bases__)
                if not any(issubclass(base, Node) for base in node_bases if isinstance(base, type)):
                    bases.append(Node)
                return type(node_or_func.__name__, tuple(bases), namespace)
            else:
                base_node = generate_node(
                    func=node_or_func,
                    subnodes=tuple(get_nodes(node_or_func).values()),
                )
                return type(to_pascal_case(node_or_func.__name__), (base_node,), namespace)

        return inner if x is None else inner(x)


@typing.dataclass_transform(kw_only_default=True)
class FactoryNode(Node, abc.ABC):
    node = "factory"

    @classmethod
    @abc.abstractmethod
    def compose(cls, *args, **kwargs) -> ComposeResult:
        pass

    def __new__(cls, **context: typing.Any) -> type[typing.Self]:
        namespace = dict(**cls.__dict__)
        namespace.pop("__new__", None)
        return type(cls.__name__, (cls,), namespace | context)  # type: ignore


@typing.dataclass_transform()
class DataNode(Node, abc.ABC):
    node = "data"

    @classmethod
    @abc.abstractmethod
    def compose(cls, *args, **kwargs) -> ComposeResult[typing.Self]:
        pass


@scalar_node(scope=NodeScope.PER_CALL)
class Name:
    @classmethod
    def compose(cls) -> str: ...


class GlobalNode[Value](Node):
    scope = NodeScope.GLOBAL

    @classmethod
    def set(cls, value: Value, /) -> None:
        setattr(cls, "_value", value)

    @typing.overload
    @classmethod
    def get(cls) -> Value: ...

    @typing.overload
    @classmethod
    def get[Default](cls, *, default: Default) -> Value | Default: ...

    @classmethod
    def get(cls, **kwargs: typing.Any) -> typing.Any:
        sentinel = object()
        default = kwargs.pop("default", sentinel)
        return getattr(cls, "_value") if default is sentinel else getattr(cls, "_value", default)

    @classmethod
    def unset(cls) -> None:
        if hasattr(cls, "_value"):
            delattr(cls, "_value")


__all__ = (
    "Composable",
    "ComposeError",
    "DataNode",
    "FactoryNode",
    "GlobalNode",
    "IsNode",
    "Name",
    "Node",
    "NodeImpersonation",
    "NodeProto",
    "NodeType",
    "as_node",
    "get_nodes",
    "is_node",
    "scalar_node",
    "unwrap_node",
)



import typing_extensions as typing

from telegrinder.api.api import API, HTTPClient
from telegrinder.bot.dispatch import dispatch as dp
from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.polling import polling as pg
from telegrinder.bot.polling.abc import ABCPolling
from telegrinder.modules import logger
from telegrinder.tools.loop_wrapper import ABCLoopWrapper
from telegrinder.tools.loop_wrapper import loop_wrapper as lw

Dispatch = typing.TypeVar("Dispatch", bound=ABCDispatch, default=dp.Dispatch[HTTPClient])
Polling = typing.TypeVar("Polling", bound=ABCPolling, default=pg.Polling[HTTPClient])
LoopWrapper = typing.TypeVar("LoopWrapper", bound=ABCLoopWrapper, default=lw.LoopWrapper)


class Telegrinder(typing.Generic[HTTPClient, Dispatch, Polling, LoopWrapper]):
    def __init__(
        self,
        api: API[HTTPClient],
        *,
        dispatch: Dispatch | None = None,
        polling: Polling | None = None,
        loop_wrapper: LoopWrapper | None = None,
    ) -> None:
        self.api = api
        self.dispatch = typing.cast(Dispatch, dispatch or dp.Dispatch())
        self.polling = typing.cast(Polling, polling or pg.Polling(api))
        self.loop_wrapper = typing.cast(LoopWrapper, loop_wrapper or lw.LoopWrapper())

    def __repr__(self) -> str:
        return "<{}: api={!r}, dispatch={!r}, polling={!r}, loop_wrapper={!r}>".format(
            self.__class__.__name__,
            self.api,
            self.dispatch,
            self.polling,
            self.loop_wrapper,
        )

    @property
    def on(self) -> Dispatch:
        return self.dispatch

    async def reset_webhook(self) -> None:
        if not (await self.api.get_webhook_info()).unwrap().url:
            return
        await self.api.delete_webhook()

    async def run_polling(
        self,
        *,
        offset: int = 0,
        skip_updates: bool = False,
    ) -> typing.NoReturn:
        async def polling() -> typing.NoReturn:
            if skip_updates:
                logger.debug("Dropping pending updates")
                await self.reset_webhook()
                await self.api.delete_webhook(drop_pending_updates=True)
            self.polling.offset = offset

            async for updates in self.polling.listen():
                for update in updates:
                    logger.debug(
                        "Received update (update_id={}, update_type={!r})",
                        update.update_id,
                        update.update_type.name,
                    )
                    self.loop_wrapper.add_task(self.dispatch.feed(update, self.api))

        if self.loop_wrapper.is_running:
            await polling()
        else:
            self.loop_wrapper.add_task(polling())
            self.loop_wrapper.run_event_loop()

    def run_forever(self, *, offset: int = 0, skip_updates: bool = False) -> typing.NoReturn:
        logger.debug("Running blocking polling (id={})", self.api.id)
        self.loop_wrapper.add_task(self.run_polling(offset=offset, skip_updates=skip_updates))
        self.loop_wrapper.run_event_loop()


__all__ = ("Telegrinder",)



from __future__ import annotations

import io
import ssl
import typing

import certifi

import telegrinder.msgspec_json as json
from telegrinder.client.abc import ABCClient

if typing.TYPE_CHECKING:
    from aiosonic import Connection, HTTPClient, HttpResponse, MultipartForm, Proxy, TCPConnector, Timeouts

type Data = dict[str, typing.Any] | MultipartForm
type Response = HttpResponse

AIOSONIC_OBJECTS = (
    "Connection",
    "HTTPClient",
    "HttpResponse",
    "Proxy",
    "TCPConnector",
    "Timeouts",
)


def init_aiosonic_module(client_class: type[ABCClient[typing.Any]], /) -> None:
    try:
        import aiosonic
        import aiosonic.exceptions
    except ImportError:
        raise ImportError(
            "Module 'aiosonic' is not installed. You can install as follows: pip install aiosonic"
        ) from None

    globalns = globals()
    for name in AIOSONIC_OBJECTS:
        globalns.setdefault(name, getattr(aiosonic, name))

    if "MultipartForm" not in globalns:
        globalns["MultipartForm"] = type("MultiPartForm", (_MultipartForm, aiosonic.MultipartForm), {})

    if not client_class.CONNECTION_TIMEOUT_ERRORS:
        client_class.CONNECTION_TIMEOUT_ERRORS = (
            aiosonic.exceptions.ConnectTimeout,
            aiosonic.exceptions.RequestTimeout,
            TimeoutError,
        )

    if not client_class.CLIENT_CONNECTION_ERRORS:
        client_class.CLIENT_CONNECTION_ERRORS = (
            aiosonic.exceptions.ConnectionDisconnected,
            aiosonic.exceptions.ConnectionPoolAcquireTimeout,
        )


class _MultipartForm(MultipartForm if typing.TYPE_CHECKING else object):
    async def _generate_chunks(self) -> typing.AsyncGenerator[bytes, None]:
        for field in self.fields:
            yield (f"--{self.boundary}\r\n").encode()

            if isinstance(field[1], io.IOBase):
                yield (
                    "Content-Type: application/octet-stream\r\n"
                    "Content-Disposition: form-data; " + f'name="{field[0]}"; filename="{field[2]}"\r\n\r\n'
                ).encode()

                async for data in self._read_file(field[1]):
                    yield data + b"\r\n"

                field[1].close()
            else:
                yield (
                    "Content-Type: text/plain; charset=utf-8\r\n"
                    f'Content-Disposition: form-data; name="{field[0]}"\r\n\r\n'
                ).encode()
                yield field[1].encode() + b"\r\n"

        yield (f"--{self.boundary}--").encode()

    async def get_body_size(self) -> tuple[bytes, int]:
        if not self.fields:
            return b"", 0
        return await super().get_body_size()

    def get_headers(self, size: int | None = None) -> dict[str, str]:
        if not self.fields:
            return {"Content-Type": "application/x-www-form-urlencoded"}
        return super().get_headers(size)


class AiosonicClient(ABCClient["MultipartForm"]):
    """HTTP client based on `aiosonic` module."""

    def __init_subclass__(cls, *args: typing.Any, **kwargs: typing.Any) -> None:
        init_aiosonic_module(cls)
        return super().__init_subclass__(*args, **kwargs)

    def __init__(
        self,
        *,
        verify_ssl: bool = True,
        tpc_pool_size: int = 25,
        tpc_timeouts: Timeouts | None = None,
        proxy: Proxy | None = None,
        conn_max_requests: int = 100,
        use_dns_cache: bool = True,
        handle_cookies: bool = False,
        ttl_dns_cache: int = 10000,
        **kwargs: typing.Any,
    ) -> None:
        init_aiosonic_module(self.__class__)
        self.ssl = ssl.create_default_context(cafile=certifi.where())
        self.proxy = proxy
        self.verify_ssl = verify_ssl
        self.handle_cookies = handle_cookies
        self.tpc_timeouts = tpc_timeouts or Timeouts()
        self.tcp_connector = TCPConnector(
            pool_size=tpc_pool_size,
            timeouts=self.tpc_timeouts,
            connection_cls=Connection,
            conn_max_requests=conn_max_requests,
            use_dns_cache=use_dns_cache,
            ttl_dns_cache=ttl_dns_cache,
            **kwargs,
        )
        self.client: HTTPClient | None = None

    def __repr__(self) -> str:
        return "<{}: proxy={!r}, tpc_timeouts={!r}, tcp_connector={!r}, client={!r}>".format(
            self.__class__.__name__,
            self.proxy,
            self.tpc_timeouts,
            self.tcp_connector,
            self.client,
        )

    @classmethod
    def multipart_form_factory(cls) -> MultipartForm:
        return MultipartForm()

    async def request_raw(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> Response:
        if self.client is None:
            self.client = HTTPClient(
                connector=self.tcp_connector,
                handle_cookies=self.handle_cookies,
                verify_ssl=self.verify_ssl,
                proxy=self.proxy,
            )

        return await self.client.request(
            url=url,
            method=method,
            data=data,
            json_serializer=json.dumps,
            ssl=self.ssl,
            **kwargs,
        )

    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> str:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.text()

    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> dict[str, typing.Any]:
        return json.loads(await self.request_content(url, method, data, **kwargs))

    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return response.body

    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.content()

    async def close(self) -> None:
        if self.client is not None:
            await self.client.connector.cleanup()
            self.client = None


__all__ = ("AiosonicClient",)



import typing

from telegrinder.node.base import IsNode, Node


class ContainerNode(Node):
    linked_nodes: typing.ClassVar[list[IsNode]]
    composer: typing.Callable[..., typing.Awaitable[typing.Any]]

    @classmethod
    async def compose(cls, **kw: typing.Any) -> typing.Any:
        subnodes = cls.get_subnodes().keys()
        return await cls.composer(*tuple(t[1] for t in sorted(kw.items(), key=lambda t: t[0]) if t[0] in subnodes))

    @classmethod
    def get_subnodes(cls) -> dict[str, IsNode]:
        subnodes = getattr(cls, "subnodes", None)
        if subnodes is None:
            subnodes_dct = {f"node_{i}": node_t for i, node_t in enumerate(cls.linked_nodes)}
            setattr(cls, "subnodes", subnodes_dct)
            return subnodes_dct
        return subnodes

    @classmethod
    def link_nodes(
        cls,
        linked_nodes: list[IsNode],
        composer: typing.Callable[..., typing.Awaitable[typing.Any]],
    ) -> type["ContainerNode"]:
        return type(cls.__name__, (cls,), {"linked_nodes": linked_nodes, "composer": classmethod(composer)})


__all__ = ("ContainerNode",)



import typing
from abc import ABC, abstractmethod

from telegrinder.bot.cute_types.base import BaseCute

if typing.TYPE_CHECKING:
    from telegrinder.api import API
    from telegrinder.bot.dispatch.view.abc import ABCStateView


class ABCScenario[Event: BaseCute](ABC):
    @abstractmethod
    def wait(self, api: "API", view: "ABCStateView[Event]") -> typing.Any:
        pass


__all__ = ("ABCScenario",)



import typing

from fntypes.option import Nothing, Some

from telegrinder.api.api import API
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware, run_middleware
from telegrinder.bot.dispatch.return_manager.abc import ABCReturnManager
from telegrinder.model import Model
from telegrinder.modules import logger
from telegrinder.node.composer import CONTEXT_STORE_NODES_KEY, NodeScope, compose_nodes
from telegrinder.tools.adapter.abc import run_adapter
from telegrinder.tools.i18n.abc import I18nEnum
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.handler.abc import ABCHandler
    from telegrinder.bot.rules.abc import ABCRule


async def process_inner[Event: Model](
    api: API,
    event: Event,
    raw_event: Update,
    ctx: Context,
    middlewares: list[ABCMiddleware[Event]],
    handlers: list["ABCHandler[Event]"],
    return_manager: ABCReturnManager[Event] | None = None,
) -> bool:
    logger.debug("Processing {!r}...", event.__class__.__name__)
    ctx[CONTEXT_STORE_NODES_KEY] = {}  # For per-event shared nodes

    logger.debug("Run pre middlewares...")
    for m in middlewares:
        result = await run_middleware(m.pre, api, event, raw_event=raw_event, ctx=ctx, adapter=m.adapter)
        logger.debug("Middleware {!r} returned: {!r}", m, result)
        if result is False:
            return False

    found = False
    responses = []
    ctx_copy = ctx.copy()

    for handler in handlers:
        adapted_event = event

        if await handler.check(api, raw_event, ctx):
            if handler.adapter is not None:
                match await run_adapter(handler.adapter, api, raw_event, ctx):
                    case Some(value):
                        adapted_event = value
                    case Nothing():
                        continue

            found = True
            logger.debug("Handler {!r} matched, run...", handler)
            response = await handler.run(api, adapted_event, ctx)
            logger.debug("Handler {!r} returned: {!r}", handler, response)
            responses.append(response)

            if return_manager is not None:
                await return_manager.run(response, event, ctx)
            if handler.final:
                break

        ctx = ctx_copy

    logger.debug("Run post middlewares...")
    ctx.set("responses", responses)

    for m in middlewares:
        await run_middleware(
            m.post,
            api,
            event,
            raw_event=raw_event,
            ctx=ctx,
            adapter=m.adapter,
        )

    for session in ctx.get(CONTEXT_STORE_NODES_KEY, {}).values():
        await session.close(scopes=(NodeScope.PER_EVENT,))

    logger.debug(
        "{} handlers, returns {!r}",
        "No found" if not found else "Found",
        found,
    )
    return found


async def check_rule(
    api: API,
    rule: "ABCRule",
    update: Update,
    ctx: Context,
) -> bool:
    """Checks requirements, adapts update.
    Returns check result.
    """
    update_cute = None if not isinstance(update, UpdateCute) else update

    # Running adapter
    match await run_adapter(rule.adapter, api, update, ctx):
        case Some(val):
            adapted_value = val
        case Nothing():
            return False

    # Preparing update
    if isinstance(adapted_value, UpdateCute):
        update_cute = adapted_value
    elif isinstance(adapted_val := ctx.get(rule.adapter.ADAPTED_VALUE_KEY or ""), UpdateCute):
        update_cute = adapted_val
    else:
        update_cute = UpdateCute.from_update(update, bound_api=api)

    # Running subrules to fetch requirements
    ctx_copy = ctx.copy()
    for requirement in rule.requires:
        if not await check_rule(api, requirement, update_cute, ctx_copy):
            return False

    # Translating translatable rules
    if I18nEnum.I18N in ctx:
        rule = await rule.translate(ctx[I18nEnum.I18N])

    ctx |= ctx_copy

    # Composing required nodes
    nodes = rule.required_nodes
    node_col = None
    if nodes:
        result = await compose_nodes(nodes, ctx, data={Update: update, API: api})
        if not result:
            logger.debug(f"Cannot compose nodes for rule, error: {str(result.error)}")
            return False
        node_col = result.value

    # Running check
    result = await rule.bounding_check(ctx, adapted_value=adapted_value, node_col=node_col)

    # Closing node sessions if there are any
    if node_col is not None:
        await node_col.close_all()

    return result


__all__ = ("check_rule", "process_inner")
